"""This module contains functionality for modifying existing ExodusII meshes.

Part of exodus_helper 1.0: Copyright 2023 Sandia Corporation
This Software is released under the BSD license detailed in the file
`license.txt` in the top-level directory"""

# --------------------------------------------------------------------------- #

from itertools import permutations
import os
import sys

import netCDF4
import numpy as np

from .core import Exodus
from .topology import RectangularPrism


# --------------------------------------------------------------------------- #

IDXS_EDGES_4 = [[0, 1], [1, 2], [0, 2], [0, 3], [1, 3], [2, 3]]


# --------------------------------------------------------------------------- #

def convert_tet4_tet10(filename_from, filename_to=None, **kwargs):
    """Convert a 4-node tetrahedral mesh into a 10-node tetrahedral mesh by
    adding mid-edge nodes to existing edges.

    Args:
        filename_from (str): Name of a .g file created in association with a
        tet4 mesh.

        filename_to (str): Name of the resultant .g file for the tet10 mesh.

    Returns:
        mesh_to (): The converted mesh.
    """

    # Extract the tet4 mesh
    mesh_from = Exodus(filename_from)

    # Initiate the tet10 connectivity from the tet4 connectivity.
    ### NOTE: NODE NUMBERS IN SIERRA START AT 1 ###
    num_elements = mesh_from.get_num_elems()
    connectivity_to = np.zeros((num_elements, 10), dtype=int)
    connectivity_from = mesh_from.get_elem_connectivity_full()[:]
    connectivity_to[:, :4] = connectivity_from

    # Get coordinates of vertex nodes
    coords_from = np.stack(mesh_from.get_coords()).T
    num_nodes_from = len(coords_from)

    # Doing something, ns =14 and ss=6
    dims = mesh_from.dataset.dimensions
    num_ns = dims['num_ns_global'].size
    num_ss = dims['num_ss_global'].size

    # Getting sides for the things above. it comes out to 242 for each instance
    # of ss
    num_side_sss = [dims[f'num_side_ss{n}'].size for n in range(1, num_ss + 1)]

    # Initiate dictionary that gives the node number for the mid-node of an
    # edge
    pairs_elem_all = {}

    # Initiate coordinates for tet10 mesh
    coords_to = coords_from.tolist()

    # Loop over the elements in the tet4 mesh
    for idx_elem, ids_node in enumerate(connectivity_from):

        # Create the pairs of nodes which define an edge for this element. We
        # sort them in ascending order otherwise the logic-gate below results
        # in redundant nodes:
        pairs_elem = [tuple(sorted(sublist)) for sublist in
                      [tuple(ids_node[i]) for i in IDXS_EDGES_4]]

        # Loop over the edges of the element
        for idx_pair, pair in enumerate(pairs_elem):

            # If the edge has already been considered, consult the dictionary
            if pair in pairs_elem_all:
                id_node_new = pairs_elem_all[pair]
            else:
                # Get the node number for the new node at the mid-point
                id_node_new = num_nodes_from + len(pairs_elem_all) + 1
                pairs_elem_all[pair] = id_node_new

                # Get coordinates for new mid-edge node. Since sierra node
                # numbers start at one, subtract one for averaging
                idxs_old = [p - 1 for p in pair]
                coords_to.append(np.mean(coords_from[idxs_old, :], axis=0))

            # Add this node to the connectivity of the tet10 mesh
            connectivity_to[idx_elem, 4 + idx_pair] = id_node_new

    ids_node_set = mesh_from.get_node_set_ids()
    sets_nodes = []
    for id_node_set in ids_node_set:
        ids_node = mesh_from.get_node_set_nodes(id_node_set).tolist()
        pairs = permutations(ids_node, 2)
        for pair in pairs:
            id_new = pairs_elem_all.get(pair, pairs_elem_all.get(pair[::-1]))
            if id_new not in ids_node and id_new is not None:
                ids_node.append(id_new)
        sets_nodes.append(ids_node)

    num_nod_nss = [len(s) for s in sets_nodes]

    ids_blk = mesh_from.get_elem_blk_ids()
    define_maps = True
    nums_elems = [mesh_from.get_num_elems_in_blk(i) for i in ids_blk]
    nums_nodes = [10 for i in ids_blk]
    nums_attrs = [0 for i in ids_blk]
    elem_types = ['TETRA10' for i in ids_blk]

    dict_attrs, dict_dimensions, dict_variables = mesh_from._get_dicts_netcdf()

    if filename_to is None:
        filename_to = '_tet10'.join(os.path.splitext(filename_from))
    mesh_to = Exodus(
        filename_to,
        mode='w',
        num_info=mesh_from.get_num_info_records() + 1,
        num_dim=mesh_from.numDim,
        num_nodes=len(coords_to),
        num_elem=num_elements,
        num_el_blk=len(ids_blk),
        num_node_sets=num_ns,
        num_side_sets=num_ss,
        num_nod_nss=num_nod_nss,
        num_side_sss=num_side_sss,
        title=kwargs.get('title', mesh_from.get_title()),
        **kwargs)

    mesh_to.dataset.createDimension('time_step', size=None)

    mesh_to.put_coords(*np.array(coords_to).T)
    mesh_to.put_concat_elem_blk(
        ids_blk, elem_types, nums_elems, nums_nodes, nums_attrs, define_maps)
    variables = mesh_to.dataset.variables
    for i in ids_blk:
        idxs_elem = mesh_from.get_idxs_elem_in_blk(i)
        variables[f'connect{i}'][:] = connectivity_to[idxs_elem]

    ids_side_set = mesh_from.get_side_set_ids()
    for id_side_set in ids_side_set:
        ids_elem, ids_side = mesh_from.get_side_set(id_side_set)
        num_sides = mesh_from.get_num_faces_in_side_set(id_side_set)
        mesh_to.put_side_set_params(id_side_set, num_sides)
        mesh_to.put_side_set(id_side_set, ids_elem, ids_side)

    for id_node_set, ids_node in zip(ids_node_set, sets_nodes):
        mesh_to.put_node_set_params(id_node_set, len(ids_node))
        mesh_to.put_node_set(id_node_set, ids_node)

    return mesh_to


def create_sets_canonical(filename):
    """Create a mesh with node/side sets consistent with a canonical
    rectangular prism.

    Args:
        filename (str): Name of a .g file created in association with a mesh.

    Raises:
        AssertionError: The filename must correspond to an existing .g file.
    """

    assert os.path.isfile(filename)
    root, ext = os.path.splitext(filename)

    mesh = RectangularPrism(filename)
    dataset_from = mesh.dataset

    dataset_to = netCDF4.Dataset(
        root + '_canonical' + ext, mode='w', format=dataset_from.data_model)

    for attr in dataset_from.ncattrs():
        dataset_to.setncattr(attr, dataset_from.getncattr(attr))

    dimensions = dataset_from.dimensions
    variables = dataset_from.variables

    for k in list(dimensions.keys()):
        if k.startswith('num_side_ss'):
            dimensions.pop(k)
        elif k.startswith('num_nod_ns'):
            dimensions.pop(k)

    dimensions.pop('num_node_sets', None)
    dimensions.pop('num_ns_global', None)
    dimensions.pop('num_side_sets', None)
    dimensions.pop('num_ss_global', None)

    for k, v in dimensions.items():
        dataset_to.createDimension(k, size=v.size)

    dataset_to.createDimension('num_node_sets', size=14)
    dataset_to.createDimension('num_ns_global', size=14)
    nodesets = {}

    for i in range(1, 7):
        nodesets[i] = mesh.get_nodes_on_surface(i)
        dataset_to.createDimension(f'num_nod_ns{i}', size=len(nodesets[i]))
    for i in range(7, 15):
        dataset_to.createDimension(f'num_nod_ns{i}', size=1)

    dataset_to.createDimension('num_side_sets', size=6)
    dataset_to.createDimension('num_ss_global', size=6)

    for k in list(variables.keys()):
        if k.startswith('side_ss'):
            variables.pop(k)
        elif k.startswith('elem_ss'):
            variables.pop(k)
        elif k.startswith('dist_fact_ns'):
            variables.pop(k)
        elif k.startswith('node_ns'):
            variables.pop(k)

    variables.pop('ns_prop1', None)
    variables.pop('ns_status', None)
    variables.pop('ns_names', None)
    variables.pop('ss_prop1', None)
    variables.pop('ss_status', None)
    variables.pop('ss_names', None)

    for k, v in variables.items():
        if '_FillValue' in v.ncattrs():
            v2 = dataset_to.createVariable(
                k, v.dtype, dimensions=v.dimensions,
                fill_value=v[:].fill_value)
        else:
            v2 = dataset_to.createVariable(
                k, v.dtype, dimensions=v.dimensions)
        v2.setncatts({a: v.getncattr(a) for a in v.ncattrs()})
        v2[:] = v[:]

    if dataset_from.getncattr('int64_status') == 0:
        int_type = np.dtype('int32')
    else:
        int_type = np.dtype('int64')

    for k, nodes in nodesets.items():
        variable = dataset_to.createVariable(
            f'node_ns{k}', int_type, dimensions=f'num_nod_ns{k}')
        variable[:] = nodes

    i = 6
    for x in range(1, 3):
        for y in range(3, 5):
            for z in range(5, 7):
                i += 1
                intersect = np.intersect1d(nodesets[x], nodesets[y])
                nodes = np.intersect1d(intersect, nodesets[z])
                assert len(nodes) == 1
                variable = dataset_to.createVariable(
                    f'node_ns{i}', int_type, dimensions=f'num_nod_ns{i}')
                variable[0] = nodes[0]

    v = dataset_to.createVariable(
        'ns_prop1', int_type, dimensions='num_node_sets')
    v[:] = np.arange(dataset_to.dimensions['num_node_sets'].size) + 1
    v.setncattr('name', 'ID')

    v = dataset_to.createVariable(
        'ns_status', int_type, dimensions='num_node_sets')
    v[:] = [1] * dataset_to.dimensions['num_node_sets'].size

    dataset_to.createVariable(
        'ns_names', np.dtype('S1'), dimensions=('num_node_sets', 'len_name'))

    for i in range(1, 7):
        elems, sides = mesh.get_elements_sides_on_surface(i)
        assert len(elems) == len(sides)
        dataset_to.createDimension(f'num_side_ss{i}', size=len(elems))
        variable = dataset_to.createVariable(
            f'elem_ss{i}', int_type, dimensions=f'num_side_ss{i}')
        variable[:] = elems
        variable = dataset_to.createVariable(
            f'side_ss{i}', int_type, dimensions=f'num_side_ss{i}')
        variable[:] = sides
    mesh.close()

    v = dataset_to.createVariable(
        'ss_prop1', v.datatype, dimensions='num_side_sets')
    v[:] = np.arange(dataset_to.dimensions['num_side_sets'].size) + 1
    v.setncattr('name', 'ID')

    v = dataset_to.createVariable(
        'ss_status', v.datatype, dimensions='num_side_sets')
    v[:] = [1] * dataset_to.dimensions['num_side_sets'].size

    dataset_to.createVariable(
        'ss_names', np.dtype('S1'), dimensions=('num_side_sets', 'len_name'))

    dataset_to.close()


def scale_mesh(filename, scale=(1., 1., 1.)):
    """Scale the dimensions of a mesh by multiplying the coordinate values of
    the mesh by the scaling weights.

    Args:
        filename (str): Name of a .g file created in association with a mesh.
        scale (tuple(float), optional): A list of scaling weights.
            Defaults to (1., 1., 1.).

    Raises:
        AssertionError: The filename must correspond to an existing .g file.
    """
    assert os.path.isfile(filename)
    root, ext = os.path.splitext(filename)

    mesh = Exodus(filename)
    dataset_from = mesh.dataset

    dataset_to = netCDF4.Dataset(
        root + '_scaled' + ext, mode='w', format=dataset_from.data_model)

    for attr in dataset_from.ncattrs():
        dataset_to.setncattr(attr, dataset_from.getncattr(attr))

    dimensions = dataset_from.dimensions
    for k, v in dimensions.items():
        dataset_to.createDimension(k, size=v.size)

    variables = dataset_from.variables
    for k, v in variables.items():
        attrs = v.ncattrs()
        fargs = {}
        if '_FillValue' in attrs:
            fargs['fill_value'] = v.getncattr('_FillValue')
            attrs.remove('_FillValue')
        variable = dataset_to.createVariable(
            k, v.datatype, dimensions=v.dimensions, **fargs)
        variable[:] = v[:]
        for attr in attrs:
            variable.setncattr(attr, v.getncattr(attr))

    mesh.close()
    dataset_to.variables['coordx'][:] *= scale[0]
    dataset_to.variables['coordy'][:] *= scale[1]
    dataset_to.variables['coordz'][:] *= scale[2]
    dataset_to.close()


# --------------------------------------------------------------------------- #

if __name__ == '__main__':
    create_sets_canonical(sys.argv[1])

# --------------------------------------------------------------------------- #



"""This module contains functionality related to producing graphical
representations of finite element meshes.

Part of exodus_helper 1.0: Copyright 2023 Sandia Corporation
This Software is released under the BSD license detailed in the file
`license.txt` in the top-level directory"""

# --------------------------------------------------------------------------- #

import numpy as np
from scipy.spatial import Delaunay
from scipy.spatial.distance import cdist

from .topology import get_dims_surface


# --------------------------------------------------------------------------- #

def render_mesh(mesh, surface, sample_ratio=0.5, return_coords=False):
    """Render the surface of a mesh as a binary image.

    Given a surface, numbered as such: 1:x-, 2:x+, 3:y-, 4:y+, 5:z-, 6:z+
    find all the elements associated with that surface, and create a binary
    image with 1's within the mesh, and 0's outside.

    Args:
        mesh (`Exodus`): A finite element mesh.
        surface (`int`): The surface to render.
        sample_ratio(`float`, optional): ratio of pixel size to element size

    Returns:
        A `numpy.ndarray` of binary values indicating whether each pixel falls
        within the mesh boundaries.
    """

    # Get all elements on the given surface and get their locations
    elems = mesh.get_elems_on_surface_sorted(surface)
    centroids = mesh.get_elem_centroids()[elems - 1]

    # Determine the spatial extent of the data
    dims = get_dims_surface(surface)
    x_min = np.min(centroids[:, dims[0]])
    l_x = np.max(centroids[:, dims[0]]) - x_min
    y_min = np.min(centroids[:, dims[1]])
    l_y = np.max(centroids[:, dims[1]]) - y_min

    # Compute the resolution of the rendered image
    d_a = l_x * l_y / len(centroids)
    d_x = np.sqrt(d_a) * sample_ratio
    num_x = int(l_x / d_x) + 1 if d_x > 0. else 1
    num_y = int(l_y / d_x) + 1 if d_x > 0. else 1

    # Create a rectangular grid of pixels
    (x_g, y_g) = np.meshgrid(
        x_min + d_x * np.arange(num_x), y_min + d_x * np.arange(num_y))
    coords_pixel = np.column_stack((x_g.flat, y_g.flat))

    # For each centroid, how far is each pixel
    distances = cdist(centroids[:, dims], coords_pixel)

    # Find the nodes associated with each element and triangulate
    nodes = mesh.get_elem_connectivity_full()[elems - 1]
    coords = np.column_stack(mesh.get_coords())
    hulls = [Delaunay(coords[np.ix_(n - 1, dims)]) for n in nodes]

    # For each pixel, get the closest neighboring elements
    neighborhoods = np.argsort(distances.T, axis=1)[:, :5]

    # Start with a blank image and fill in points that reside within an element
    colors = np.zeros((num_x * num_y))
    for idx, neighbors in enumerate(neighborhoods):
        for neighbor in neighbors:
            if hulls[neighbor].find_simplex(coords_pixel[idx]) >= 0:
                colors[idx] = 1
                break

    if return_coords:
        return (colors.reshape((num_y, num_x)), coords_pixel)

    return colors.reshape((num_y, num_x))


def map_points_to_elements(mesh, points):
    centroids = mesh.get_elem_centroids()
    distances = cdist(centroids, points)
    neighborhoods = np.argsort(distances.T, axis=1)
    nodes = mesh.get_elem_connectivity_full()
    coords = np.column_stack(mesh.get_coords())
    hulls = [Delaunay(coords[n - 1]) for n in nodes]
    parents = -np.ones(len(points), dtype=int)
    for idx, neighbors in enumerate(neighborhoods):
        for neighbor in neighbors:
            if hulls[neighbor].find_simplex(points[idx]) >= 0:
                parents[idx] = neighbor
                break
        if parents[idx] == -1:
            raise ValueError
    return parents

# --------------------------------------------------------------------------- #



"""This module contains the RectangularPrism class, derived from the core
Exodus class, and external functions related to querying objects of this type.

Part of exodus_helper 1.0: Copyright 2023 Sandia Corporation
This Software is released under the BSD license detailed in the file
`license.txt` in the top-level directory"""

# --------------------------------------------------------------------------- #

import re

import numpy as np

from .core import CONNECTIVITY_SIDES
from .core import NUM_NODES_SIDE
from .core import Exodus


# --------------------------------------------------------------------------- #

# exodus_helper rectangular prism surface id canonical order
# 1:x-, 2:x+, 3:y-, 4:y+, 5:z-, 6:z+
PICK_SURFACE = {
    1: (np.min, 0),
    2: (np.max, 0),
    3: (np.min, 1),
    4: (np.max, 1),
    5: (np.min, 2),
    6: (np.max, 2)}


# --------------------------------------------------------------------------- #

class RectangularPrism(Exodus):
    """Specialization of the Exodus class for rectangular prism geometries.
        This geometry simplifies eg identifying elements on a surface"""

    def __init__(
            self, filename, shape=(1, 1, 1), resolution=(1., 1., 1.),
            num_attr=0, mode='r', **kwargs):
        """Create an `Exodus` mesh object for a rectangular prism.

        Args:
            filename (str): File containing mesh data.
            shape (tuple(int), optional): The shape of the prism as a length-3
                `tuple`. Defaults to (1, 1, 1).
            resolution (tuple(float), optional): The (spatially invariant)
                resolution of each element. Defaults to (1., 1., 1.).
            num_attr (int, optional): The number of attributes for each element
                block. Defaults to 0.
            **kwargs: Variable length keyword arguments to pass to
                Exodus.__init__.

        Returns:
            Exodus: An `Exodus` mesh for a rectangular prism.
        """

        if mode == 'r':
            super().__init__(filename, mode='r', **kwargs)
            return

        num_elements = np.prod(shape)
        num_nodes_x = shape[0] + 1
        num_nodes_y = shape[1] + 1
        num_nodes_z = shape[2] + 1
        num_nodes = num_nodes_x * num_nodes_y * num_nodes_z

        idxs_x = np.arange(num_nodes_x)
        idxs_y = np.arange(num_nodes_y)
        idxs_z = np.arange(num_nodes_z)

        idxs = np.meshgrid(idxs_y, idxs_z, idxs_x)

        idxs_yg = idxs[0].flatten()
        idxs_zg = idxs[1].flatten()
        idxs_xg = idxs[2].flatten()

        coords_x = resolution[0] * idxs_xg
        coords_y = resolution[1] * idxs_yg
        coords_z = resolution[2] * idxs_zg

        # Each element has the following nodal connectivity:
        #
        #        8 ------ 7
        #       /|       /|
        #      / |      / |     z
        #     5 ------ 6  |     |  y
        #     |  4 ----|- 3     | /
        #     | /      | /       --- x
        #     |/       |/
        #     1 ------ 2

        connectivity = np.zeros((num_elements, 8), dtype=int)
        num_xy = num_nodes_x * num_nodes_y
        for idx_z in range(shape[2]):
            for idx_y in range(shape[1]):
                for idx_x in range(shape[0]):
                    i = idx_z * shape[0] * shape[1] + idx_y * shape[0] + idx_x
                    base = idx_y * num_nodes_x + idx_x + idx_z * num_xy + 1
                    connectivity[i, :] = [
                        base, base + 1,
                        base + num_nodes_x + 1,
                        base + num_nodes_x,
                        base + num_xy,
                        base + 1 + num_xy,
                        base + num_nodes_x + 1 + num_xy,
                        base + num_nodes_x + num_xy]

        num_node_nss = [
            num_nodes_y * num_nodes_z,
            num_nodes_y * num_nodes_z,
            num_nodes_x * num_nodes_z,
            num_nodes_x * num_nodes_z,
            num_nodes_x * num_nodes_y,
            num_nodes_x * num_nodes_y,
            1, 1, 1, 1, 1, 1, 1, 1]

        num_side_sss = [
            shape[1] * shape[2],
            shape[1] * shape[2],
            shape[0] * shape[2],
            shape[0] * shape[2],
            shape[0] * shape[1],
            shape[0] * shape[1]]

        elem_blk_info = kwargs.get(
            'elem_blk_info',
            ([1], ['HEX8'], [num_elements], [8], [num_attr], True))

        num_el_blk = len(elem_blk_info[0])

        super().__init__(
            filename,
            mode='w',
            num_info=1,
            time_step=1,
            num_dim=3,
            num_nodes=num_nodes,
            num_elem=num_elements,
            num_el_blk=num_el_blk,
            num_node_sets=14,
            num_side_sets=6,
            num_processors=1,
            num_procs_file=1,
            num_nod_nss=num_node_nss,
            num_side_sss=num_side_sss,
            title='rectangular_prism',
            **kwargs)

        self.put_coords(coords_x, coords_y, coords_z)
        self.put_concat_elem_blk(*elem_blk_info)
        starts = np.cumsum(np.concatenate(([0], elem_blk_info[2])))
        for idx in range(num_el_blk):
            id_blk = elem_blk_info[0][idx]
            connectivity_blk = connectivity[starts[idx]: starts[idx + 1]]
            self.put_elem_connectivity(id_blk, connectivity_blk)

        node_set_nodes = []
        side_set_elements = []
        side_set_sides = []
        for id_surface in range(1, 7):
            node_set_nodes.append(self.get_nodes_on_surface(id_surface))
            elems, sides = self.get_elems_sides_on_surface(id_surface)
            side_set_elements.append(elems)
            side_set_sides.append(sides)
        for x in range(2):
            for y in range(2, 4):
                for z in range(4, 6):
                    nodes_xy = np.intersect1d(
                        node_set_nodes[x], node_set_nodes[y])
                    intersect = np.intersect1d(nodes_xy, node_set_nodes[z])
                    node_set_nodes.append(intersect)

        for idx, nodes in enumerate(node_set_nodes):
            self.put_node_set_params(idx + 1, len(nodes), numSetDistFacts=1)
            self.put_node_set(idx + 1, nodes)

        for idx, element in enumerate(side_set_elements):
            num_sides = len(side_set_sides[idx])
            self.put_side_set_params(idx + 1, num_sides, numSetDistFacts=4)
            self.put_side_set(idx + 1, element, side_set_sides[idx])

        self.put_elem_id_map(range(1, num_elements + 1))
        self.put_node_id_map(range(1, num_nodes + 1))

    @property
    def resolution(self):
        return self.get_resolution()

    @property
    def shape(self):
        """tuple(int, int, int): The shape of the mesh."""
        return self.get_shape()

    def get_elems_on_surface(self, surface):
        return self.get_elems_sides_on_surface(surface)[0]

    def get_elements_on_surface(self, surface):
        return self.get_elems_on_surface(surface)

    def get_elems_on_surface_sorted(self, surface):
        elems = self.get_elems_on_surface(surface)
        centroids = self.get_elem_centroids()[elems - 1]
        dims = get_dims_surface(surface)
        idxs_sort = np.lexsort([centroids[:, d] for d in dims])
        return elems[idxs_sort]

    def get_elements_on_surface_sorted(self, surface):
        return self.get_elems_on_surface_sorted(surface)

    def get_elems_sides_on_surface(self, surface):
        nodes_surface = self.get_nodes_on_surface(surface)
        idxs_start = self.get_idxs_elem_start_blk()
        elems = []
        sides = []
        for id_blk, idx_start in zip(self.get_elem_blk_ids(), idxs_start):
            elem_type, _, num_nodes, _ = self.get_elem_blk_info(id_blk)
            num_nodes = NUM_NODES_SIDE[elem_type]
            connectivity_info = self.get_elem_connectivity(id_blk)
            connectivity = connectivity_info[0].reshape(connectivity_info[1:])
            idxs_surf_elem = get_idxs_surface_elem(
                nodes_surface, connectivity, idx_start)
            faces_surface = [
                i for i in idxs_surf_elem if len(i[1]) == num_nodes]
            connectivity_sides = CONNECTIVITY_SIDES[
                re.sub(r'\d+', '', elem_type)]
            num_check = len(list(connectivity_sides.keys())[0])
            for face in faces_surface:
                elems.append(face[0])
                sides.append(connectivity_sides[tuple(face[1][:num_check])])
        return np.array(elems, dtype=int), np.array(sides, dtype=int)

    def get_elements_sides_on_surface(self, surface):
        return self.get_elems_sides_on_surface(surface)

    def get_elems_sorted(self, order='xyz'):
        if isinstance(order, str):
            dims = {'x': 0, 'y': 1, 'z': 2}
            order = [dims[o] for o in order]
        centroids = self.get_elem_centroids()
        return np.lexsort([centroids[:, o] for o in order]) + 1

    def get_elements_sorted(self, order='xyz'):
        return self.get_elems_sorted(order=order)

    def get_nodes_on_surface(self, surface):
        coordinates = np.column_stack(self.get_coords())
        side, dim = PICK_SURFACE[surface]
        nodes = np.where(
            np.isclose(side(coordinates[:, dim]), coordinates[:, dim]))[0] + 1
        return nodes

    def get_patches_on_surface(self, surface):
        nodes_surface = self.get_nodes_on_surface(surface)
        coordinates = np.column_stack(self.get_coords())
        idxs_start = self.get_idxs_elem_start_blk()
        patches = []
        for id_blk, idx_start in zip(self.get_elem_blk_ids(), idxs_start):
            elem_type, _, num_nodes, _ = self.get_elem_blk_info(id_blk)
            num_nodes = NUM_NODES_SIDE[elem_type]
            connectivity_info = self.get_elem_connectivity(id_blk)
            connectivity = connectivity_info[0].reshape(connectivity_info[1:])
            nodes_surface_elem = get_nodes_surface_elem(
                nodes_surface, connectivity, idx_start)
            faces_surface = [
                n for n in nodes_surface_elem if len(n[1]) == num_nodes]
            connectivity_sides = CONNECTIVITY_SIDES[
                re.sub(r'\d+', '', elem_type)]
            num_check = len(list(connectivity_sides.keys())[0])
            for face in faces_surface:
                patches.append(np.array(
                    [coordinates[f - 1] for f in face[1][:num_check]]))
        return patches

    def get_resolution(self):
        conn = self.get_elem_connectivity_full()
        coords = np.column_stack(self.get_coords())
        return np.max(np.abs(np.diff(coords[conn[0] - 1], axis=0)), axis=0)

    def get_shape(self):
        num_xs = self.numElem // self.get_elements_on_surface(1).size
        num_ys = self.numElem // self.get_elements_on_surface(3).size
        num_zs = self.numElem // self.get_elements_on_surface(5).size
        assert num_xs * num_ys * num_zs == self.numElem
        return (num_xs, num_ys, num_zs)

    def get_shape_surface(self, surface):
        dims = get_dims_surface(surface)
        shape = self.get_shape()
        return tuple(shape[d] for d in dims)

    def get_sides_on_surface(self, surface):
        return self.get_elems_sides_on_surface(surface)[1]


def get_dims_surface(surface):
    """Get the indices of the dimensions that span a surface.

    The surface ID's for a rectangular prism are:

        1:x-, 2:x+, 3:y-, 4:y+, 5:z-, 6:z+

    Args:
        surface (int): Surface ID.

    Returns:
        A `numpy.ndarray` of dimension indicies (ints).
    """
    dim = PICK_SURFACE[surface][1]
    return np.sort([(dim - 1) % 3, (dim + 1) % 3])


def get_idxs_surface_elem(nodes, connectivity, idx_start=0):
    shift = idx_start + 1
    itr = enumerate(connectivity)
    return [(i + shift, np.where(np.in1d(c, nodes))[0] + 1) for i, c in itr]


def get_nodes_surface_elem(nodes, connectivity, idx_start=0):
    idxs = get_idxs_surface_elem(nodes, connectivity)
    return [(i[0] + idx_start, connectivity[i[0] - 1, i[1] - 1]) for i in idxs]

# --------------------------------------------------------------------------- #



"""This module contains functions for computing quantities (eg volumes) for
finite elements

Part of exodus_helper 1.0: Copyright 2023 Sandia Corporation
This Software is released under the BSD license detailed in the file
`license.txt` in the top-level directory"""

# --------------------------------------------------------------------------- #

import numpy as np


# --------------------------------------------------------------------------- #

def calculate_volume_element(coords, shape_functions='trilinear'):
    """Calculate the volume of an element from a list of nodal coordinates.

    Args:
        coords (list(list(float))): A length-8 (HEX) list of nodal
            coordinate values.
        shape_functions (str, optional): The name of the shape function used to
            interpolate the element's volume from a list of discrete nodal
            coordinate values. Defaults to `trilinear`.

    Returns:
        float: The volume of the element.

    Raises:
        Exception: `trilinear` is currently the only supported shape function.
    """
    coords = np.array(coords)
    if shape_functions == 'trilinear':
        v_12 = coords[0, 2] * coords[1, 1] * coords[2, 0] \
            - coords[0, 1] * coords[1, 2] * coords[2, 0] \
            - coords[0, 2] * coords[1, 0] * coords[2, 1] \
            + coords[0, 0] * coords[1, 2] * coords[2, 1] \
            + coords[0, 1] * coords[1, 0] * coords[2, 2] \
            - coords[0, 0] * coords[1, 1] * coords[2, 2] \
            + coords[0, 2] * coords[1, 1] * coords[3, 0] \
            - coords[0, 1] * coords[1, 2] * coords[3, 0] \
            + coords[0, 2] * coords[2, 1] * coords[3, 0] \
            + coords[1, 2] * coords[2, 1] * coords[3, 0] \
            - coords[0, 1] * coords[2, 2] * coords[3, 0] \
            - coords[1, 1] * coords[2, 2] * coords[3, 0] \
            - coords[0, 2] * coords[1, 0] * coords[3, 1] \
            + coords[0, 0] * coords[1, 2] * coords[3, 1] \
            - coords[0, 2] * coords[2, 0] * coords[3, 1] \
            - coords[1, 2] * coords[2, 0] * coords[3, 1] \
            + coords[0, 0] * coords[2, 2] * coords[3, 1] \
            + coords[1, 0] * coords[2, 2] * coords[3, 1] \
            + coords[0, 1] * coords[1, 0] * coords[3, 2] \
            - coords[0, 0] * coords[1, 1] * coords[3, 2] \
            + coords[0, 1] * coords[2, 0] * coords[3, 2] \
            + coords[1, 1] * coords[2, 0] * coords[3, 2] \
            - coords[0, 0] * coords[2, 1] * coords[3, 2] \
            - coords[1, 0] * coords[2, 1] * coords[3, 2] \
            - coords[0, 2] * coords[1, 1] * coords[4, 0] \
            + coords[0, 1] * coords[1, 2] * coords[4, 0] \
            + coords[0, 2] * coords[3, 1] * coords[4, 0] \
            - coords[0, 1] * coords[3, 2] * coords[4, 0] \
            + coords[0, 2] * coords[1, 0] * coords[4, 1] \
            - coords[0, 0] * coords[1, 2] * coords[4, 1] \
            - coords[0, 2] * coords[3, 0] * coords[4, 1] \
            + coords[0, 0] * coords[3, 2] * coords[4, 1] \
            - coords[0, 1] * coords[1, 0] * coords[4, 2] \
            + coords[0, 0] * coords[1, 1] * coords[4, 2] \
            + coords[0, 1] * coords[3, 0] * coords[4, 2] \
            - coords[0, 0] * coords[3, 1] * coords[4, 2] \
            - coords[0, 2] * coords[1, 1] * coords[5, 0] \
            + coords[0, 1] * coords[1, 2] * coords[5, 0] \
            - coords[1, 2] * coords[2, 1] * coords[5, 0] \
            + coords[1, 1] * coords[2, 2] * coords[5, 0] \
            + coords[0, 2] * coords[4, 1] * coords[5, 0] \
            + coords[1, 2] * coords[4, 1] * coords[5, 0] \
            - coords[0, 1] * coords[4, 2] * coords[5, 0] \
            - coords[1, 1] * coords[4, 2] * coords[5, 0] \
            + coords[0, 2] * coords[1, 0] * coords[5, 1] \
            - coords[0, 0] * coords[1, 2] * coords[5, 1] \
            + coords[1, 2] * coords[2, 0] * coords[5, 1] \
            - coords[1, 0] * coords[2, 2] * coords[5, 1] \
            - coords[0, 2] * coords[4, 0] * coords[5, 1] \
            - coords[1, 2] * coords[4, 0] * coords[5, 1] \
            + coords[0, 0] * coords[4, 2] * coords[5, 1] \
            + coords[1, 0] * coords[4, 2] * coords[5, 1] \
            - coords[0, 1] * coords[1, 0] * coords[5, 2] \
            + coords[0, 0] * coords[1, 1] * coords[5, 2] \
            - coords[1, 1] * coords[2, 0] * coords[5, 2] \
            + coords[1, 0] * coords[2, 1] * coords[5, 2] \
            + coords[0, 1] * coords[4, 0] * coords[5, 2] \
            + coords[1, 1] * coords[4, 0] * coords[5, 2] \
            - coords[0, 0] * coords[4, 1] * coords[5, 2] \
            - coords[1, 0] * coords[4, 1] * coords[5, 2] \
            - coords[1, 2] * coords[2, 1] * coords[6, 0] \
            + coords[1, 1] * coords[2, 2] * coords[6, 0] \
            - coords[2, 2] * coords[3, 1] * coords[6, 0] \
            + coords[2, 1] * coords[3, 2] * coords[6, 0] \
            + coords[1, 2] * coords[5, 1] * coords[6, 0] \
            + coords[2, 2] * coords[5, 1] * coords[6, 0] \
            - coords[4, 2] * coords[5, 1] * coords[6, 0] \
            - coords[1, 1] * coords[5, 2] * coords[6, 0] \
            - coords[2, 1] * coords[5, 2] * coords[6, 0] \
            + coords[4, 1] * coords[5, 2] * coords[6, 0] \
            + coords[1, 2] * coords[2, 0] * coords[6, 1] \
            - coords[1, 0] * coords[2, 2] * coords[6, 1] \
            + coords[2, 2] * coords[3, 0] * coords[6, 1] \
            - coords[2, 0] * coords[3, 2] * coords[6, 1] \
            - coords[1, 2] * coords[5, 0] * coords[6, 1] \
            - coords[2, 2] * coords[5, 0] * coords[6, 1] \
            + coords[4, 2] * coords[5, 0] * coords[6, 1] \
            + coords[1, 0] * coords[5, 2] * coords[6, 1] \
            + coords[2, 0] * coords[5, 2] * coords[6, 1] \
            - coords[4, 0] * coords[5, 2] * coords[6, 1] \
            - coords[1, 1] * coords[2, 0] * coords[6, 2] \
            + coords[1, 0] * coords[2, 1] * coords[6, 2] \
            - coords[2, 1] * coords[3, 0] * coords[6, 2] \
            + coords[2, 0] * coords[3, 1] * coords[6, 2] \
            + coords[1, 1] * coords[5, 0] * coords[6, 2] \
            + coords[2, 1] * coords[5, 0] * coords[6, 2] \
            - coords[4, 1] * coords[5, 0] * coords[6, 2] \
            - coords[1, 0] * coords[5, 1] * coords[6, 2] \
            - coords[2, 0] * coords[5, 1] * coords[6, 2] \
            + coords[4, 0] * coords[5, 1] * coords[6, 2] \
            + coords[0, 2] * coords[3, 1] * coords[7, 0] \
            - coords[2, 2] * coords[3, 1] * coords[7, 0] \
            - coords[0, 1] * coords[3, 2] * coords[7, 0] \
            + coords[2, 1] * coords[3, 2] * coords[7, 0] \
            - coords[0, 2] * coords[4, 1] * coords[7, 0] \
            - coords[3, 2] * coords[4, 1] * coords[7, 0] \
            + coords[0, 1] * coords[4, 2] * coords[7, 0] \
            + coords[3, 1] * coords[4, 2] * coords[7, 0] \
            - coords[4, 2] * coords[5, 1] * coords[7, 0] \
            + coords[4, 1] * coords[5, 2] * coords[7, 0] \
            + coords[2, 2] * coords[6, 1] * coords[7, 0] \
            + coords[3, 2] * coords[6, 1] * coords[7, 0] \
            - coords[4, 2] * coords[6, 1] * coords[7, 0] \
            - coords[5, 2] * coords[6, 1] * coords[7, 0] \
            - coords[2, 1] * coords[6, 2] * coords[7, 0] \
            - coords[3, 1] * coords[6, 2] * coords[7, 0] \
            + coords[4, 1] * coords[6, 2] * coords[7, 0] \
            + coords[5, 1] * coords[6, 2] * coords[7, 0] \
            - coords[0, 2] * coords[3, 0] * coords[7, 1] \
            + coords[2, 2] * coords[3, 0] * coords[7, 1] \
            + coords[0, 0] * coords[3, 2] * coords[7, 1] \
            - coords[2, 0] * coords[3, 2] * coords[7, 1] \
            + coords[0, 2] * coords[4, 0] * coords[7, 1] \
            + coords[3, 2] * coords[4, 0] * coords[7, 1] \
            - coords[0, 0] * coords[4, 2] * coords[7, 1] \
            - coords[3, 0] * coords[4, 2] * coords[7, 1] \
            + coords[4, 2] * coords[5, 0] * coords[7, 1] \
            - coords[4, 0] * coords[5, 2] * coords[7, 1] \
            - coords[2, 2] * coords[6, 0] * coords[7, 1] \
            - coords[3, 2] * coords[6, 0] * coords[7, 1] \
            + coords[4, 2] * coords[6, 0] * coords[7, 1] \
            + coords[5, 2] * coords[6, 0] * coords[7, 1] \
            + coords[2, 0] * coords[6, 2] * coords[7, 1] \
            + coords[3, 0] * coords[6, 2] * coords[7, 1] \
            - coords[4, 0] * coords[6, 2] * coords[7, 1] \
            - coords[5, 0] * coords[6, 2] * coords[7, 1] \
            + coords[0, 1] * coords[3, 0] * coords[7, 2] \
            - coords[2, 1] * coords[3, 0] * coords[7, 2] \
            - coords[0, 0] * coords[3, 1] * coords[7, 2] \
            + coords[2, 0] * coords[3, 1] * coords[7, 2] \
            - coords[0, 1] * coords[4, 0] * coords[7, 2] \
            - coords[3, 1] * coords[4, 0] * coords[7, 2] \
            + coords[0, 0] * coords[4, 1] * coords[7, 2] \
            + coords[3, 0] * coords[4, 1] * coords[7, 2] \
            - coords[4, 1] * coords[5, 0] * coords[7, 2] \
            + coords[4, 0] * coords[5, 1] * coords[7, 2] \
            + coords[2, 1] * coords[6, 0] * coords[7, 2] \
            + coords[3, 1] * coords[6, 0] * coords[7, 2] \
            - coords[4, 1] * coords[6, 0] * coords[7, 2] \
            - coords[5, 1] * coords[6, 0] * coords[7, 2] \
            - coords[2, 0] * coords[6, 1] * coords[7, 2] \
            - coords[3, 0] * coords[6, 1] * coords[7, 2] \
            + coords[4, 0] * coords[6, 1] * coords[7, 2] \
            + coords[5, 0] * coords[6, 1] * coords[7, 2]
    else:
        raise ValueError('Only shape_functions=\'trilinear\' is supported')
    return v_12 / 12.


def calculate_volume_hexahedron(coords):
    """Calculate the volume of a hexahedral mesh.

    Args:
        coords (list(list(float))): A length-8 (HEX) list of nodal
            coordinate values.

    Returns:
        numpy.float64: The volume of the hexahedron.

    Raises:
        AssertionError: The coordinates are not compatible with right-hand
            positive winding.
    """
    centroid = np.mean(coords, axis=0)
    coordinates = np.row_stack((centroid, coords))
    ts = np.array([
        [1, 5, 2], [6, 2, 5], [2, 6, 3], [7, 3, 6], [3, 7, 4], [8, 4, 7],
        [1, 4, 5], [8, 5, 4], [1, 2, 4], [3, 4, 2], [6, 5, 7], [8, 7, 5]],
        dtype=int)

    volume = 0.
    for i in range(12):
        edge_1 = np.array(
            coordinates[ts[i, 1]]) - np.array(coordinates[ts[i, 0]])
        edge_2 = np.array(
            coordinates[ts[i, 2]]) - np.array(coordinates[ts[i, 0]])
        edge_3 = np.array(coordinates[0]) - np.array(coordinates[ts[i, 0]])
        tm = np.column_stack((edge_1, edge_2, edge_3))
        test = np.linalg.det(tm) >= 0.
        assert test, 'Algorithm assumes right-hand positive winding'
        volume += np.linalg.det(tm) / 6

    return volume


def calculate_volumes_element(mesh, id_blk=None):
    """Calculate the volume of each element in a specified element block within
    a given mesh.

    Args:
        mesh (Exodus): The `Exodus` instance of a mesh.
        id_blk (int): Element block ID.

    Returns:
        numpy.ndarray(float): An array of element volumes ordered by element ID
    """
    if id_blk:
        connectivity, num_elems, num_nodes = mesh.get_elem_connectivity(id_blk)
        connectivity = connectivity.reshape((num_elems, num_nodes))
    else:
        connectivity = mesh.get_elem_connectivity_full()
    coords_node = np.column_stack(mesh.get_coords())
    volumes_element = np.array(
        [calculate_volume_element(coords_node[ns - 1]) for ns in connectivity])
    return volumes_element


def calculate_volumes_block(mesh):
    """Calculate the voluemes of each element block in a mesh.

    Args:
        mesh (Exodus): The `Exodus` instance of a mesh.

    Returns:
        numpy.ndarray(float): An array of element volumes ordered by element
        block ID.
    """
    ids_blk = mesh.get_elem_blk_ids()
    volumes_blk = np.zeros(len(ids_blk))
    for idx, id_blk in enumerate(ids_blk):
        volumes_blk[idx] = np.sum(calculate_volumes_element(mesh, id_blk))
    return volumes_blk

# --------------------------------------------------------------------------- #
